/*----------------------------------------------------------*
    krslib.c
    KRSプロトコル用API V1.1
                                       Last Edit '13 11/07
   Copyright (c) 2013 BestTechnology CO.,LTD.
 *----------------------------------------------------------*/

/*----------------------------------------------------------------------
  【プログラム概要】
    WindowsAPIをベースにKRSシリーズの基本的な機能への送受信ルーチン。
    ICS3.0/3.5対応のアクチュエータで実装されていないメモリマップを仮想化している。

  【注意事項】
    COMポート番号が提供されるKRS用TLL I/Fが対象。

  【ドキュメント・その他】
    Serial Communications in Win32:
      http://msdn.microsoft.com/en-us/library/ms810467.aspx
    Serial HOWTO in Linux
      http://www.linux.org/docs/ldp/howto/Serial-HOWTO.html
    ICSコマンドリファレンス:
      http://kondo-robot.com/sys/archives/category/download/command-reference
  ----------------------------------------------------------------------*/

#ifdef _WIN32
#define   __MAKE_LIB__
#include  <tchar.h>
#include  <process.h>
#else
#include  <fcntl.h>
#include  <termios.h>
#include  <unistd.h>
#include  <errno.h>
#include  <sys/ioctl.h>
#include  <sys/time.h>
#include  <linux/serial.h>
#endif

#include  <stdio.h>
#include  <string.h>
#include  <stdlib.h>
#include  "krslib.h"
#include  "krsmemmap.h"

#define MAX_COMPORT         (256)

#define DXMAXLENGTH         (93)
#define MAX_SERVOID         (0x1f)
#define MAX_POS             (12000)
#define MIN_POS             (3000)

// 仮想メモリマップ
typedef struct {
  uint8_t   id;                 // ID
  uint8_t   torque_enable;      // 制御のON/OFF
  uint16_t  goal_position;      // 指令位置
  uint8_t   goal_speed;         // 指令速度
  uint8_t   stretch;            // 指令ストレッチ
  uint16_t  present_position;   // 現在位置
  int8_t    present_current;    // 現在電流
  int8_t    present_temp;       // 現在温度 [degC]
} __attribute__((packed)) TVirtualMemMap;

// 通信管理用構造体
typedef struct {
  char              name[30];   // ポート名
  uint32_t          baudrate;   // 通信速度
  uint32_t          offsettime; // タイムアウトのオフセット時間
  HANDLE            hComm;      // デバイスハンドル
  CRITICAL_SECTION  tpcs_s;     // 共通API排他処理用
  CRITICAL_SECTION  tpcs_m;     // 汎用API排他処理用
  TVirtualMemMap    vmemmap[MAX_SERVOID + 1];   // 仮想メモリマップ
#ifndef _WIN32
  fd_set            rfds;
#endif
} TComInfo;

static CRITICAL_SECTION Tpcs;                   // DLL内排他処理用クリティカルセクション
static TDeviceID        ComInfo[MAX_COMPORT];   // 通信ポート管理テーブル
#ifdef _WIN32
static LARGE_INTEGER    sysfreq;                // マイクロ秒計測
#endif
static bool             EchoCancel = true;      // エコーキャンセル

//------------------------------------------------------------------------------
#ifdef _WIN32
bool APIENTRY DllMain (HINSTANCE hInst, uint32_t reason, LPVOID reserved) {
  int i;

  switch (reason) {
    case DLL_PROCESS_ATTACH:
      InitializeCriticalSection (&Tpcs);
      QueryPerformanceFrequency (&sysfreq);
      for (i = 0; i < MAX_COMPORT; i++) ComInfo[i] = 0;
      break;

    case DLL_PROCESS_DETACH:
      for (i = 0; i < MAX_COMPORT; i++) {
        if (ComInfo[i]) KRS_ClosePort (ComInfo[i]);
      }
      DeleteCriticalSection (&Tpcs);
      break;

    case DLL_THREAD_ATTACH:
      break;

    case DLL_THREAD_DETACH:
      break;
  }
  return  true;
}


bool APIENTRY DllEntryPoint (HINSTANCE hInst, uint32_t reason, LPVOID reserved) {
  return DllMain (hInst, reason, reserved);
}

#else
// for Linux
static void __attribute__ ((constructor)) attach (void) {
  int         i;
  pthread_mutex_init (&Tpcs, NULL);
  for (i = 0; i < MAX_COMPORT; i++) ComInfo[i] = 0;
}

static void __attribute__ ((destructor)) detach (void) {
  int         i;
  for (i = 0; i < MAX_COMPORT; i++) {
    if (ComInfo[i]) KRS_ClosePort (ComInfo[i]);
  }
  pthread_mutex_destroy (&Tpcs);
}
#endif

/*----------------------------------------------------------------------------
    double GetQueryPerformanceCounter (void)
  ----------------------------------------------------------------------------
    DESCRIPTION:
        ミリ秒起動時間計
  ----------------------------------------------------------------------------*/
double KRSAPIDLL GetQueryPerformanceCounter (void) {
#ifdef _WIN32
  LARGE_INTEGER now;

  QueryPerformanceCounter (&now);
  return ((double)now.QuadPart / (double)sysfreq.QuadPart) * (double)1000.0;
#else
  struct timeval nowtime;
  if (gettimeofday (&nowtime, NULL) == 0) {
    return ((double)(nowtime.tv_sec) + (double)(nowtime.tv_usec) * 0.000001) * (double)1000.0;
  }
  return 0;
#endif
}

#ifndef _WIN32
//------------------------------------------------------------------------------
// long getAliveBaudrate (long baud)
//------------------------------------------------------------------------------
//  DESCRIPTION:
//    Linux用通信速度の取得
//  PARAMS:
//    long            baud        ボーレート
//  RETURNS:
//    -1              未サポートのボーレート
//    -1以外          Linux内で使用するボーレートの値
//
//  Linux専用関数
//
long getAliveBaudrate (long baud) {
  typedef struct {
    long  baud;
    long  val;
  } st_baud;
  st_baud baudrate_t [] = {
    {50, B50}, {75, B75}, {110, B110}, {134, B134}, {150, B150}, {200, B200}, {300, B300},
    {600, B600}, {1200, B1200}, {1800, B1800}, {2400, B2400}, {4800, B4800}, {9600, B9600},
    {19200, B19200}, {38400, B38400}, // 以降POSIX.1
    {57600, B57600}, {115200, B115200}, {230400, B230400}, {460800, B460800}, {500000, B500000},
    {576000, B576000}, {921600, B921600}, {1000000, B1000000}, {1152000, B1152000},
    {1500000, B1500000}, {2000000, B2000000}, {2500000, B2500000}, {3000000, B3000000},
    {3500000, B3500000}, {4000000, B4000000}
  };
  long result = -1;
  int cnt = sizeof (baudrate_t) / sizeof (st_baud);
  int i;
  for (i = 0; i < cnt; i++) {
    if (baud == baudrate_t[i].baud) {
      result = baudrate_t[i].val;
      break;
    }
  }
  return result;
}
#endif

//------------------------------------------------------------------------------
// bool KRS_Active (TDeviceID dvid)
//------------------------------------------------------------------------------
//  DESCRIPTION:
//    通信ポートが開いているかどうか
//  PARAMS:
//    TDeviceID       dvid        デバイスID
//  RETURNS:
//    true            正常
//    false           エラー
//
bool KRSAPIDLL KRS_Active (TDeviceID dvid) {
#ifdef _WIN32
  DCB           cDcb;
#else
  struct termios term;
#endif
  bool          result = false;

  if (dvid != 0) {
#ifdef _WIN32
    result = GetCommState (((TComInfo *)dvid)->hComm, &cDcb);
#else
    result = (tcgetattr (((TComInfo *)dvid)->hComm, &term) >= 0);
#endif
  }
  return result;
}

//------------------------------------------------------------------------------
//  bool KRS_ClosePort (TDeviceID dvid)
//------------------------------------------------------------------------------
//  DESCRIPTION:
//    通信ポートを閉じる
//  PARAMS:
//    TDeviceID      dvid        デバイスID
//  RETURNS:
//    true            成功
//    false           失敗
//
bool KRSAPIDLL KRS_ClosePort (TDeviceID dvid) {
  int         i;
  TComInfo    *pInfo = (TComInfo *)dvid;
  bool        result = false;

  EnterCriticalSection (&Tpcs);

  if (dvid != 0) {
    for (i = 0; i < MAX_COMPORT; i++) {
      if (ComInfo[i] == dvid) {
        ComInfo[i] = 0;
        break;
      }
    }
    if (i < MAX_COMPORT) {
#ifdef _WIN32
      if (pInfo->hComm != INVALID_HANDLE_VALUE) {
        PurgeComm (pInfo->hComm, (PURGE_TXABORT | PURGE_RXABORT | PURGE_TXCLEAR | PURGE_RXCLEAR));
        CloseHandle (pInfo->hComm);
      }
#else
      if (pInfo->hComm != INVALID_HANDLE_VALUE) {
        tcflush (pInfo->hComm, TCIOFLUSH);
        close (pInfo->hComm);
      }
#endif

      DeleteCriticalSection (&(pInfo->tpcs_s));
      DeleteCriticalSection (&(pInfo->tpcs_m));

      free (pInfo);
      result = true;
    }
  }

  LeaveCriticalSection (&Tpcs);

  return result;
}

void KRSAPIDLL KRS_SetEchoCancel (bool becho) {
  EchoCancel = becho;
}

//------------------------------------------------------------------------------
//  TDeviceID KRS_OpenPort (char *name, uint32_t baud)
//------------------------------------------------------------------------------
//  DESCRIPTION:
//    通信ポートを開く
//  PARAMS:
//    char            *name       通信ポートを指定
//    uint32_t        baud        ボーレートを指定
//  RETURNS:
//    TDeviceID       NULLでない場合：Device ID
//                    NULLの場合は失敗
//
TDeviceID KRSAPIDLL KRS_OpenPort (char *name, uint32_t baud) {
  int           i, index;
  bool          errflag = false;
  TComInfo      *pInfo = NULL;
  TDeviceID     result = 0;
#ifdef _WIN32
  DCB           cDcb;
  COMMTIMEOUTS  cTimeout = {0, 0, 0, 0, 0};
#else
  struct termios       term;
  struct serial_struct serial;
  long   lbaud;
#endif

  if (name == NULL) return 0;

  for (i = 0; i < MAX_COMPORT; i++) {
    if (ComInfo[i]) {
      if (!strcmp (((TComInfo *)ComInfo[i])->name, name)) {
        KRS_ClosePort (ComInfo[i]);
        break;
      }
    }
  }

  EnterCriticalSection (&Tpcs);

  for (i = 0; i < MAX_COMPORT; i++) {
    if (ComInfo[i] == 0) {
      index = i;
      break;
    }
  }
  if (i < MAX_COMPORT) {
    pInfo = (TComInfo *)malloc (sizeof (TComInfo));
    for (i = 0; i < MAX_SERVOID + 1; i++) pInfo->vmemmap[i].id = i;
    if (pInfo != NULL) {
      ComInfo[index] = (TDeviceID)pInfo;
      memset (pInfo, 0, sizeof(TComInfo));

      strncpy (pInfo->name, name, 30);
      pInfo->hComm = INVALID_HANDLE_VALUE;

      // デバイスオープン
#ifdef _WIN32
      memset (&cDcb,0,sizeof(cDcb));
      cDcb.BaudRate           = baud;
      cDcb.ByteSize           = 8;
      cDcb.fParity            = true;
      cDcb.Parity             = EVENPARITY;
      cDcb.StopBits           = ONESTOPBIT;
      cDcb.fOutX              = false;
      cDcb.fInX               = false;
      cDcb.fTXContinueOnXoff  = true;
      cDcb.XonLim             = 512;
      cDcb.XoffLim            = 512;
      cDcb.XonChar            = 0x11;
      cDcb.XoffChar           = 0x13;
      cDcb.fOutxDsrFlow       = false;
      cDcb.fOutxCtsFlow       = false;
      cDcb.fRtsControl        = RTS_CONTROL_DISABLE;
      cDcb.fDtrControl        = DTR_CONTROL_ENABLE;
      cDcb.fErrorChar         = 0;
      cDcb.fBinary            = true;
      cDcb.fNull              = 0;
      cDcb.fAbortOnError      = false;
      cDcb.wReserved          = 0;
      cDcb.EofChar            = 0x03;
      cDcb.EvtChar            = 0;

      pInfo->hComm = CreateFileA (name, (GENERIC_READ | GENERIC_WRITE), 0, NULL, OPEN_EXISTING, 0, NULL);
      if ((pInfo->hComm != INVALID_HANDLE_VALUE) && (pInfo->hComm != NULL)) {
#else
      pInfo->hComm = open (name, O_RDWR | O_NOCTTY);
      if (pInfo->hComm != INVALID_HANDLE_VALUE) {
#endif

        // 諸設定
#ifdef _WIN32
        if (!SetCommTimeouts (pInfo->hComm, &cTimeout)) errflag = true;
        if (!SetCommMask (pInfo->hComm, 0)) errflag = true;
        if (!SetCommState (pInfo->hComm, &cDcb)) errflag = true;
        PurgeComm (pInfo->hComm, (PURGE_TXCLEAR | PURGE_RXCLEAR));
#else
        if (!KRS_SetBaudrate ((TDeviceID)pInfo, baud)) errflag = true;
        FD_ZERO (&pInfo->rfds);
        FD_SET (pInfo->hComm, &pInfo->rfds);
#endif
        if (!errflag) {
          InitializeCriticalSection (&(pInfo->tpcs_s));
          InitializeCriticalSection (&(pInfo->tpcs_m));
          pInfo->baudrate = baud;
          pInfo->offsettime = 20;
          result = (TDeviceID)pInfo;
        }
      }
    }
  }

  LeaveCriticalSection (&Tpcs);

  if ((pInfo != NULL) && result == 0) KRS_ClosePort ((TDeviceID)pInfo);
  return result;
}

//------------------------------------------------------------------------------
//  bool KRS_SetBaudrate (TDeviceID dvid, uint32_t baud)
//------------------------------------------------------------------------------
//  DESCRIPTION:
//    通信ポートのボーレートを変更する
//  PARAMS:
//    TDeviceID       dvid        デバイスIDを指定
//    uint32_t        baud        ボーレートを指定
//  RETURNS:
//    true            成功
//    false           失敗
//
bool KRSAPIDLL KRS_SetBaudrate (TDeviceID dvid, uint32_t baud) {
  bool          result = false;
  TComInfo      *pInfo = (TComInfo *)dvid;
#ifdef _WIN32
  DCB           cDcb;
#else
  struct serial_struct serial;
  struct termios       term;
  long lbaud;
  int ierr = 0;
#endif

  if (dvid == 0) {

  } else if (pInfo->hComm == INVALID_HANDLE_VALUE) {

  } else {
    EnterCriticalSection (&(((TComInfo *)dvid)->tpcs_s));

#ifdef _WIN32
    PurgeComm (pInfo->hComm, (PURGE_TXCLEAR | PURGE_RXCLEAR));
    GetCommState (pInfo->hComm, &cDcb);
    cDcb.BaudRate = baud;
    if (SetCommState (pInfo->hComm, &cDcb)) result = true;
#else
    tcflush (pInfo->hComm, TCIOFLUSH);
    if (tcgetattr (pInfo->hComm, &term) != -1) { 
      if ((lbaud = getAliveBaudrate (baud)) == -1) {
        lbaud = B38400;
        if (ioctl (pInfo->hComm, TIOCGSERIAL, &serial) != -1) {
          serial.custom_divisor = (int)((double)serial.baud_base / (double)baud + 0.5); 
          serial.flags |= ASYNC_SPD_CUST | ASYNC_LOW_LATENCY;
          if (ioctl (pInfo->hComm, TIOCSSERIAL, &serial) == -1)  ierr = 1;
        } else ierr = 1;
      }
      if (!ierr) {
        term.c_lflag = term.c_lflag & ~(ICANON);
        term.c_cflag = CS8 | CLOCAL | CREAD | lbaud;
        term.c_cc[VMIN]  = 1; // 1 byte receive wait
        term.c_cc[VTIME] = 0; // no character timer
        if (tcsetattr (pInfo->hComm, TCSANOW, &term) != -1) result = true;
      }
    }
#endif
    pInfo->baudrate = baud;

    LeaveCriticalSection (&(pInfo->tpcs_s));
  }
  return result;
}

//------------------------------------------------------------------------------
//  void KRS_SetTimeOutOffset (TDeviceID dvid, uint32_t t)
//------------------------------------------------------------------------------
//  DESCRIPTION:
//    通信ポートのボーレートを変更する
//  PARAMS:
//    TDeviceID       dvid        デバイスIDを指定
//    uint32_t        t           タイムアウトのオフセット時間
//
void KRSAPIDLL KRS_SetTimeOutOffset (TDeviceID dvid, uint32_t t) {
  if (dvid != 0) {
    if (((TComInfo *)dvid)->hComm != INVALID_HANDLE_VALUE) {
      ((TComInfo *)dvid)->offsettime = t;
    }
  }
}

//------------------------------------------------------------------------------
//  double CalcTimeout (uint32_t baud, uint32_t num)
//------------------------------------------------------------------------------
//  DESCRIPTION:
//    タイムアウト時間の算出
//  PARAMS:
//    uint32_t        baud        ボーレート
//    uint32_t        num         バイト数
//  RETURNS:
//    理論タイムアウト時間[ms]
//
static double CalcTimeout (uint32_t baud, uint32_t num) {
  return (10000.0 * (double)num) / (double)baud;
}

//------------------------------------------------------------------------------
// マルチプラットホーム用受信ルーチン
//------------------------------------------------------------------------------
static bool readdata (TDeviceID d, uint8_t *buf, uint32_t bytetoread, uint32_t *bytesread, int tout) {
  TComInfo      *pInfo = (TComInfo *)d;

  if (pInfo->hComm == INVALID_HANDLE_VALUE) {
    return false;
  } else {
#ifdef _WIN32
    COMMTIMEOUTS timeouts = { 0, 0, tout, 0, 0 };
    SetCommTimeouts (pInfo->hComm, &timeouts);
    if (ReadFile (pInfo->hComm, buf, bytetoread, (LPDWORD)bytesread, NULL)) {
      return (bytetoread == *bytesread);
    }
    return false;
#else
    struct timeval timeouts = { tout / 1000, (tout % 1000) * 1000 };
    struct termios term;
    if (tcgetattr (pInfo->hComm, &term) == 0) {
      if (term.c_cc[VMIN] != bytetoread) {
        term.c_cc[VMIN] = bytetoread;
        tcsetattr (pInfo->hComm, TCSANOW, &term);
      }
    }
    fd_set fds;
    memcpy(&fds, &pInfo->rfds, sizeof(fd_set));
    if (select (pInfo->hComm + 1, &fds, NULL, NULL, &timeouts) == 0) {
      return false;
    } else {
      *bytesread = read (pInfo->hComm, buf, bytetoread);
      return (bytetoread == *bytesread);
    }
#endif
  }
  return false;
}

//------------------------------------------------------------------------------
//  int KRS_TxPacket (TDeviceID dvid, uint8_t id, TCommand cmd, uint8_t *param, uint32_t len, TErrorCode *err)
//------------------------------------------------------------------------------
//  DESCRIPTION:
//    コマンドの送信
//  PARAMS:
//    TDeviceID       dvid        デバイスID
//    uint8_t         id          ID
//    TCommand        cmd         コマンド
//    uint8_t         *param      送信データ(パケットの2バイト目以降)
//    uint32_t        iLength     送信データサイズ
//    TErrorCode      *err        エラーコード
//  RETURNS:
//    true            送信正常終了
//    false           エラー
//
bool KRSAPIDLL KRS_TxPacket (TDeviceID dvid, uint8_t id, TCommand cmd, uint8_t *param, uint32_t len, TErrorCode *err) {
#ifdef _WIN32
  COMSTAT       comStat;
  uint32_t      dwErrors;
#endif
  TComInfo      *pInfo = (TComInfo *)dvid;
  bool          result = false;
  uint8_t       tbuf[100], rbuf[100];
  int           i;
  uint32_t      wlen, rlen;

  if (dvid == 0) {
    if (err != NULL) *err = ERR_INVALID_DEVID;
  } else if (((TComInfo *)dvid)->hComm == INVALID_HANDLE_VALUE) {
    if (err != NULL) *err = ERR_COMM;
  } else if (id > MAX_SERVOID) {
    if (err != NULL) *err = ERR_INVALID_ID;
  } else {
    // idとcmdを元に先頭1バイトを作成
    tbuf[0] = ((cmd << 5) & 0xe0) | (id & 0x1f);
    for (i = 0; i < len; i++) tbuf[i + 1] = param[i];
#ifdef _WIN32
    ClearCommError (pInfo->hComm, (LPDWORD)&dwErrors, &comStat);
    PurgeComm (pInfo->hComm, (PURGE_RXCLEAR));
    if (WriteFile (pInfo->hComm, tbuf, len + 1, (LPDWORD)&wlen, NULL)) {
#else
    tcflush (pInfo->hComm, TCIFLUSH);
    if ((wlen = write (pInfo->hComm, tbuf, len + 1)) >= 0) {
#endif
      if (EchoCancel) {
        if (readdata (dvid, rbuf, len + 1, &rlen, 50)) {
          for (i = 0; i < len + 1; i++) if (tbuf[i] != rbuf[i]) result = false;
          result = true;
        } else result = true;
      } else result = true;
    } else if (err != NULL) *err = ERR_COMM;
  }
  return result;
}

//------------------------------------------------------------------------------
//  bool KRS_RxPacket (TDeviceID dvid, uint8_t *rdata, uint32_t rdatasize, uint32_t *rlen, uint32_t timeout, TErrorCode *err);
//------------------------------------------------------------------------------
//  DESCRIPTION:
//    応答の受信
//  PARAMS:
//    TDeviceID       dvid        デバイスIDを指定
//    uint8_t         *rdata      受信されたデータの保存先
//    uint32_t        rdatasize   受信データの保存先サイズ
//    uint32_t        *rlen       受信できたデータのバイト数の保存先
//    uint32_t        timeout     受信タイムアウトの時間をmsで指定
//    TErrorCode      *err        エラーコード
//  RETURNS:
//    true            指定バイト数の受信が指定時間内に完了 (それ以外はノーチェック)
//    false           エラー
//
bool KRSAPIDLL KRS_RxPacket (TDeviceID dvid, uint8_t *rdata, uint32_t rdatasize, uint32_t *rlen, uint32_t timeout, TErrorCode *err) {
  bool            result = false;

  if (dvid == 0) {
    if (err != NULL) *err = ERR_INVALID_DEVID;
  } else if (((TComInfo *)dvid)->hComm == INVALID_HANDLE_VALUE) {
    if (err != NULL) *err = ERR_COMM;
  } else if (rdata == NULL) {
    if (err != NULL) *err = ERR_COMM;
  } else if (rdatasize < 1) {
    if (err != NULL) *err = ERR_INVALID_PARAM;
  } else {
    if (readdata (dvid, rdata, rdatasize, rlen, timeout)) {
      result = true;
    } else if (err != NULL) *err = ERR_TIMEOUT;
  }
  return result;
}

//------------------------------------------------------------------------------
// 位置の書き込み (指定位置のコマンド送信と仮想メモリマップへの情報書き込み)
//------------------------------------------------------------------------------
static bool _WritePos (TDeviceID dvid, uint8_t id, const uint16_t pos, TErrorCode *err) {
  TComInfo  *pInfo = (TComInfo *)dvid;
  uint32_t  len;
  uint8_t   param[5], rbuf[4];
  uint8_t   compati;
  uint32_t  timeout;

  compati = (id == 0) && (pInfo->baudrate == 115200);  // ICS2.0互換判定

  param[0] = (pos >> 7) & 0x7f;
  param[1] = 0x7f & pos;
  if (KRS_TxPacket (dvid, id, CMD_WRITE_POS, param, 2, err)) {
    timeout = ((TComInfo *)dvid)->offsettime + CalcTimeout (((TComInfo *)dvid)->baudrate, 3 + 3);
    if (KRS_RxPacket (dvid, rbuf, 3, &len, timeout, err)) {
      if (rbuf[0] == ((compati ? (CMD_WRITE_POS << 5) : 0) | id)) {
        pInfo->vmemmap[id].goal_position = pos;
        pInfo->vmemmap[id].torque_enable = (pos == 0) ? 0 : 1;
        pInfo->vmemmap[id].present_position = (uint16_t)((rbuf[1] << 7) | (rbuf[2] & 0x7f));
        return true;
      } else if (err != NULL) *err |= ERR_INVALID_PARAM;
    }
  }
  return false;
}

//------------------------------------------------------------------------------
// 位置の読み込み (仮想メモリマップのトルクON/OFF状態で送信位置を適宜判断)
//------------------------------------------------------------------------------
static bool _ReadPos (TDeviceID dvid, uint8_t id, TErrorCode *err) {
  TComInfo  *pInfo = (TComInfo *)dvid;

  return _WritePos (dvid, id, pInfo->vmemmap[id].torque_enable ? pInfo->vmemmap[id].goal_position : 0, err);
}

//------------------------------------------------------------------------------
// トルクON/OFF設定 (仮想メモリマップのトルクON/OFF状態で送信位置を適宜判断)
//------------------------------------------------------------------------------
static bool _SetTorque (TDeviceID dvid, uint8_t id, bool On, TErrorCode *err) {
  TComInfo  *pInfo = (TComInfo *)dvid;

  if (On) {
    if (pInfo->vmemmap[id].torque_enable)
      return _WritePos (dvid, id, pInfo->vmemmap[id].goal_position, err);
    else if (_WritePos (dvid, id, 0, err))
      return _WritePos (dvid, id, pInfo->vmemmap[id].present_position, err);
  } else
    return _WritePos (dvid, id, 0, err);
  return false;
}

//------------------------------------------------------------------------------
// 速度取得 (スピード取得のコマンド送信と仮想メモリマップへの情報書き込み)
//------------------------------------------------------------------------------
static bool _ReadSpeed (TDeviceID dvid, uint8_t id, TErrorCode *err) {
  TComInfo  *pInfo = (TComInfo *)dvid;
  uint32_t  len;
  const uint8_t param[1] = {SC_SPEED};
  uint8_t   rbuf[4];
  uint32_t  timeout;

  if (KRS_TxPacket (dvid, id, CMD_READ_PARAM, (uint8_t *)param, 1, err)) {
    timeout = ((TComInfo *)dvid)->offsettime + CalcTimeout (((TComInfo *)dvid)->baudrate, 3 + 3);
    if (KRS_RxPacket (dvid, rbuf, 3, &len, timeout, err)) {
      if ((rbuf[0] == ((CMD_READ_PARAM & 0x3) << 5 | id) && rbuf[1] == param[0])) {
        pInfo->vmemmap[id].goal_speed = rbuf[2];
        return true;
      } else if (err != NULL) *err |= ERR_INVALID_PARAM;
    }
  }
  return false;
}

//------------------------------------------------------------------------------
// 速度設定 (スピード設定のコマンド送信と仮想メモリマップへの情報書き込み)
//------------------------------------------------------------------------------
static bool _WriteSpeed (TDeviceID dvid, uint8_t id, uint8_t spd, TErrorCode *err) {
  TComInfo  *pInfo = (TComInfo *)dvid;
  uint32_t  len;
  uint8_t   param[2] = {SC_SPEED, spd};
  uint8_t   rbuf[4];
  uint32_t  timeout;

  if (spd >=1 && spd <= 127) {
    if (KRS_TxPacket (dvid, id, CMD_WRITW_PARAM, param, 2, err)) {
      timeout = ((TComInfo *)dvid)->offsettime + CalcTimeout (((TComInfo *)dvid)->baudrate, 3 + 3);
      if (KRS_RxPacket (dvid, rbuf, 3, &len, timeout, err)) {
        if ((rbuf[0] == ((CMD_WRITW_PARAM & 0x3) << 5 | id)) && (rbuf[1] == param[0]) && (rbuf[2] == param[1])) {
          pInfo->vmemmap[id].goal_speed = spd;
          return true;
        } else if (err != NULL) *err |= ERR_INVALID_PARAM;
      }
    }
  } else if (err != NULL) *err |= ERR_KRS_RANGE;
  return false;
}

//------------------------------------------------------------------------------
// ストレッチ取得 (ストレッチ取得のコマンド送信)
//------------------------------------------------------------------------------
static bool _ReadStretch (TDeviceID dvid, uint8_t id, TErrorCode *err) {
  TComInfo  *pInfo = (TComInfo *)dvid;
  uint32_t  len;
  const uint8_t param[1] = {SC_STRETCH};
  uint8_t   rbuf[4];
  uint32_t  timeout;

  if (KRS_TxPacket (dvid, id, CMD_READ_PARAM, (uint8_t *)param, 1, err)) {
    timeout = ((TComInfo *)dvid)->offsettime + CalcTimeout (((TComInfo *)dvid)->baudrate, 2 + 3);
    if (KRS_RxPacket (dvid, rbuf, 3, &len, timeout, err)) {
      if ((rbuf[0] == ((CMD_READ_PARAM & 0x3) << 5 | id) && rbuf[1] == param[0])) {
        pInfo->vmemmap[id].stretch = rbuf[2];
        return true;
      } else if (err != NULL) *err |= ERR_INVALID_PARAM;
    }
  }
  return false;
}

//------------------------------------------------------------------------------
// ストレッチ設定 (ストレッチ設定のコマンド送信)
//------------------------------------------------------------------------------
static bool _WriteStretch (TDeviceID dvid, uint8_t id, uint8_t strc, TErrorCode *err) {
  TComInfo  *pInfo = (TComInfo *)dvid;
  uint32_t  len;
  uint8_t   param[2] = {SC_STRETCH, strc};
  uint8_t   rbuf[4];
  uint32_t  timeout;

  if (strc >=1 && strc <= 127) {
    if (KRS_TxPacket (dvid, id, CMD_WRITW_PARAM, param, 2, err)) {
      timeout = ((TComInfo *)dvid)->offsettime + CalcTimeout (((TComInfo *)dvid)->baudrate, 3 + 3);
      if (KRS_RxPacket (dvid, rbuf, 3, &len, timeout, err)) {
        if ((rbuf[0] == ((CMD_WRITW_PARAM & 0x3) << 5 | id)) && (rbuf[1] == param[0]) && (rbuf[2] == param[1])) {
          pInfo->vmemmap[id].stretch = strc;
          return true;
        } else if (err != NULL) *err |= ERR_INVALID_PARAM;
      }
    }
  } else if (err != NULL) *err |= ERR_KRS_RANGE;
  return false;
}

//------------------------------------------------------------------------------
// 現在電流取得 (電流取得のコマンド送信と仮想メモリマップへの情報書き込み)
//------------------------------------------------------------------------------
static bool _ReadCurrent (TDeviceID dvid, uint8_t id, TErrorCode *err) {
  TComInfo  *pInfo = (TComInfo *)dvid;
  uint32_t  len;
  const uint8_t param[1] = {SC_CURRENT};
  uint8_t   rbuf[4];
  uint32_t  timeout;

  if (KRS_TxPacket (dvid, id, CMD_READ_PARAM, (uint8_t *)param, 1, err)) {
    timeout = ((TComInfo *)dvid)->offsettime + CalcTimeout (((TComInfo *)dvid)->baudrate, 2 + 3);
    if (KRS_RxPacket (dvid, rbuf, 3, &len, timeout, err)) {
      if ((rbuf[0] & 0x7f) == ((CMD_READ_PARAM & 0x3) << 5 | id)) {
        pInfo->vmemmap[id].present_current = (rbuf[2] & 0x40) ? -((int)(rbuf[2] & 0x3f)) : (int)(rbuf[2] & 0x3f);
        return true;
      } else if (err != NULL) *err |= ERR_INVALID_PARAM;
    }
  }
  return false;
}

//------------------------------------------------------------------------------
// 現在温度取得 (温度取得のコマンド送信と仮想メモリマップへの情報書き込み)
//------------------------------------------------------------------------------
static bool _ReadTemperature (TDeviceID dvid, uint8_t id, TErrorCode *err) {
  TComInfo  *pInfo = (TComInfo *)dvid;
  int       t;
  uint32_t  len;
  const uint8_t param[1] = {SC_TEMPTURE};
  uint8_t   rbuf[4];
  uint32_t  timeout;

  if (KRS_TxPacket (dvid, id, CMD_READ_PARAM, (uint8_t *)param, 1, err)) {
    timeout = ((TComInfo *)dvid)->offsettime + CalcTimeout (((TComInfo *)dvid)->baudrate, 2 + 3);
    if (KRS_RxPacket (dvid, rbuf, 3, &len, timeout, err)) {
      if (rbuf[0] == ((CMD_READ_PARAM & 0x3) << 5 | id)) {
        t = (127.0 - (int)rbuf[2]) * 9.0 / 11.0 + 41.55;
        if (t < INT8_MIN) t = INT8_MIN;
        else if (t > INT8_MAX) t = INT8_MAX;
        pInfo->vmemmap[id].present_temp = (int8_t)t;
        return true;
      } else if (err != NULL) *err |= ERR_INVALID_PARAM;
    }
  }
  return false;
}

//------------------------------------------------------------------------------
//  bool KRS_ReadByteData (TDeviceID dvid, uint8_t id, uint8_t adr, uint8_t *rdata, TErrorCode *err)
//-- ----------------------------------------------------------------------------
//  DESCRIPTION:
//    任意アドレスの1バイトデータ読み出し
//  PARAMS:
//    TDeviceID       dvid        デバイスID
//    uint8_t         id          送信先IDを指定
//    uint8_t         adr         アドレスを指定
//    uint8_t         *rdatat     受信されたデータの保存先
//    TErrorCode      *err        エラーコード
//  RETURNS:
//    true            成功
//    false           エラー
//
bool KRSAPIDLL KRS_ReadByteData (TDeviceID dvid, uint8_t id, uint8_t adr, uint8_t *rdata, TErrorCode *err) {
  return KRS_ReadBlockData (dvid, id, adr, rdata, 1, err);
}

//------------------------------------------------------------------------------
//  bool KRS_ReadWordData (TDeviceID dvid, uint8_t id, uint8_t adr, uint16_t *rdata, TErrorCode *err)
//------------------------------------------------------------------------------
//  DESCRIPTION:
//    任意アドレスの1ワードデータ読み出し
//  PARAMS:
//    TDeviceID       dvid        デバイスID
//    uint8_t         id          送信先IDを指定
//    uint8_t         adr         アドレスを指定
//    uint16_t        *rdata      受信されたデータの保存先
//    TErrorCode      *err    エラーコード
//  RETURNS:
//    true            成功
//    false           エラー
//
bool KRSAPIDLL KRS_ReadWordData (TDeviceID dvid, uint8_t id, uint8_t adr, uint16_t *rdata, TErrorCode *err) {
  return KRS_ReadBlockData (dvid, id, adr, (uint8_t *)rdata, 2, err);
}

//------------------------------------------------------------------------------
//  bool KRS_WriteByteData (TDeviceID dvid, uint8_t id, uint8_t adr, uint8_t dat, TErrorCode *err)
//------------------------------------------------------------------------------
//  DESCRIPTION:
//    任意アドレスへ1バイトデータ書き込み
//  PARAMS:
//    TDeviceID       dvid        デバイスID
//    uint8_t         id          送信先IDを指定
//    uint8_t         adr         アドレスを指定
//    uint8_t         dat         送信するデータを指定
//    TErrorCode      *err        エラーコード
//  RETURNS:
//    true            成功
//    false           エラー
//
bool KRSAPIDLL KRS_WriteByteData (TDeviceID dvid, uint8_t id, uint8_t adr, uint8_t dat, TErrorCode *err) {
  return KRS_WriteBlockData (dvid, id, adr, (uint8_t *)&dat, 1, err);
}

//------------------------------------------------------------------------------
//  bool KRS_WriteWordData (TDeviceID dvid, uint8_t id, uint8_t adr, uint16_t dat, TErrorCode *err)
//------------------------------------------------------------------------------
//  DESCRIPTION:
//    任意アドレスへ1ワードデータ書き込み
//  PARAMS:
//    TDeviceID       dvid        デバイスID
//    uint8_t         id          送信先IDを指定
//    uint8_t         adr         アドレスを指定
//    uint16_t        dat         送信するデータを指定
//    TErrorCode      *err        エラーコード
//  RETURNS:
//    true            成功
//    false           エラー
//
bool KRSAPIDLL KRS_WriteWordData (TDeviceID dvid, uint8_t id, uint8_t adr, uint16_t dat, TErrorCode *err) {
  return KRS_WriteBlockData (dvid, id, adr, (uint8_t *)&dat, 2, err);
}

//------------------------------------------------------------------------------
//  bool KRS_ReadBlockData (TDeviceID dvid, uint8_t id, uint8_t adr, uint8_t *rdata, uint32_t len, TErrorCode *err)
//------------------------------------------------------------------------------
//  DESCRIPTION:
//    任意アドレスから任意バイト数のデータ読み出し
//  PARAMS:
//    TDeviceID       dvid        デバイスID
//    uint8_t         id          送信先IDを指定
//    uint8_t         adr         アドレスを指定
//    uint8_t         *rdata      受信されたデータの保存先
//    uint32_t        len         読み出すデータのサイズを指定
//    TErrorCode      *err        エラーコード
//  RETURNS:
//    true            成功
//    false           エラー
//
bool KRSAPIDLL KRS_ReadBlockData (TDeviceID dvid, uint8_t id, uint8_t adr, uint8_t *rdata, uint32_t len, TErrorCode *err) {
  bool  result = false;

  if (dvid == 0) {
    if (err != NULL) *err = ERR_INVALID_DEVID;
  } else if (((TComInfo *)dvid)->hComm == INVALID_HANDLE_VALUE) {
    if (err != NULL) *err = ERR_COMM;
  } else  if (id == BROADCASTING_ID || id > MAX_SERVOID) {
    if (err != NULL) *err = ERR_INVALID_ID;
  } else  if (len < 1 || (adr + len) > sizeof (TVirtualMemMap)) {
    if (err != NULL) *err = ERR_ILLEGAL_SIZE;
  } else {
    EnterCriticalSection (&(((TComInfo *)dvid)->tpcs_m));

    result = true;
    if (err != NULL) *err = ERR_NON;
    if (adr == ADDRESS_ID) {
      result = _ReadPos (dvid, id, err);
    }
    if (result && adr <= ADDRESS_TORQUE_ENABLE && (adr + len > ADDRESS_TORQUE_ENABLE)) ;
    if (result && adr <= ADDRESS_GOAL_POSITION && (adr + len > ADDRESS_GOAL_POSITION)) ;
    if (result && adr <= ADDRESS_GOAL_SPEED && (adr + len) > ADDRESS_GOAL_SPEED) {
      result = _ReadSpeed (dvid, id, err);
    }
    if (result && adr <= ADDRESS_STRETCH && (adr + len) > ADDRESS_STRETCH) {
      result = _ReadStretch (dvid, id, err);
    }
    if (adr <= ADDRESS_PRESENT_POSITION && (adr + len > ADDRESS_PRESENT_POSITION + 1)) {
      result = _ReadPos (dvid, id, err);
    }
    if (result && adr <= ADDRESS_PRESENT_LOAD && (adr + len) > ADDRESS_PRESENT_LOAD) {
      result = _ReadCurrent (dvid, id, err);
    }
    if (result && adr <= ADDRESS_PRESENT_TEMP && (adr + len) > ADDRESS_PRESENT_TEMP) {
      result = _ReadTemperature (dvid, id, err);
    }
    if (result) memcpy (rdata, &((char *)(&(((TComInfo *)dvid)->vmemmap[id])))[adr], len);

    LeaveCriticalSection (&(((TComInfo *)dvid)->tpcs_m));
  }
  return result;
}

//------------------------------------------------------------------------------
//  bool KRS_WriteBlockData (TDeviceID dvid, uint8_t id, uint8_t adr, uint8_t *dat, uint32_t len, TErrorCode *err)
//------------------------------------------------------------------------------
//  DESCRIPTION:
//    任意アドレスへ任意バイト数のデータ書き込み
//  PARAMS:
//    TDeviceID       dvid        デバイスID
//    uint8_t         id          送信先IDを指定
//    uint8_t         adr         アドレスを指定
//    uint8_t         *dat        送信するデータを指定
//    uint32_t        len         書き込むデータのサイズを指定
//    TErrorCode      *err        エラーコード
//  RETURNS:
//    true            成功
//    false           エラー
//
bool KRSAPIDLL KRS_WriteBlockData (TDeviceID dvid, uint8_t id, uint8_t adr, uint8_t *dat, uint32_t len, TErrorCode *err) {

  // 1軸分の処理
  bool write_block_data (uint8_t p_id) {
    bool result = false;
    uint16_t w;
    uint8_t ofs;

    if (adr <= ADDRESS_GOAL_SPEED && (adr + len) > ADDRESS_GOAL_SPEED) {
      ofs = ADDRESS_GOAL_SPEED - adr;
      result = _WriteSpeed (dvid, p_id, dat[ofs], err);
    }
    if (adr <= ADDRESS_STRETCH && (adr + len) > ADDRESS_STRETCH) {
      ofs = ADDRESS_STRETCH - adr;
      result = _WriteStretch (dvid, p_id, dat[ofs], err);
    }
    if (adr <= ADDRESS_TORQUE_ENABLE && (adr + len) > ADDRESS_TORQUE_ENABLE) {
      ofs = ADDRESS_TORQUE_ENABLE - adr;
      result = _SetTorque (dvid, p_id, (dat[ofs] != 0), err);
    }
    if (adr <= ADDRESS_GOAL_POSITION && (adr + len) > (ADDRESS_GOAL_POSITION + 1)) {
      ofs = ADDRESS_GOAL_POSITION - adr;
      w = dat[ofs + 1] << 8 | dat[ofs];
      if (w <= MAX_POS && w >= MIN_POS) {
        if (((TComInfo *)dvid)->vmemmap[id].torque_enable) result = _WritePos (dvid, p_id, w, err);
        else result = true;
      } else {
        if (err != NULL) *err = ERR_KRS_RANGE;
        result = false;
      }
    }
    return result;
  }

  bool        result = false;
  uint8_t     ids;

  if (dvid == 0) {
    if (err != NULL) *err = ERR_INVALID_DEVID;
  } else if (((TComInfo *)dvid)->hComm == INVALID_HANDLE_VALUE) {
    if (err != NULL) *err = ERR_COMM;
  } else if (len < 1 || (adr + len) > sizeof (TVirtualMemMap)) {
    if (err != NULL) *err = ERR_ILLEGAL_SIZE;
  } else if (id > MAX_SERVOID && id != BROADCASTING_ID) {
    if (err != NULL) *err = ERR_INVALID_ID;
  } else {
    EnterCriticalSection (&(((TComInfo *)dvid)->tpcs_m));

    if (err != NULL) *err = ERR_NON;
    if (id == BROADCASTING_ID) {
      for (ids = 0; ids < MAX_SERVOID; ids++) {
        if (write_block_data (ids)) {
          result = true;
        } else {
          if (err != NULL) *err |= ERR_KRS_RANGE;
          result = false;
          break;
        }
      }
      result = true;
    } else {
      if (write_block_data (id)) {
        result = true;
      } else {
        if (err != NULL) *err |= ERR_KRS_RANGE;
        result = false;
      }
    }

    LeaveCriticalSection (&(((TComInfo *)dvid)->tpcs_m));
  }
  return result;
}

//------------------------------------------------------------------------------
//  bool KRS_Ping (TDeviceID dvid, uint8_t id, TErrorCode *err)
//------------------------------------------------------------------------------
//  DESCRIPTION:
//    任意アドレスのデバイスへPINGを送信し応答を受信
//  PARAMS:
//    TDeviceID       dvid        デバイスID
//    uint8_t         id          送信先IDを指定
//    TErrorCode      *err        エラーコード
//  RETURNS:
//    true            検出
//    false           エラー
//
bool KRSAPIDLL KRS_Ping (TDeviceID dvid, uint8_t id, TErrorCode *err) {
  bool          result = false;

  if (dvid == 0) {
    if (err != NULL) *err = ERR_INVALID_DEVID;
  } else if (((TComInfo *)dvid)->hComm == INVALID_HANDLE_VALUE) {
    if (err != NULL) *err = ERR_COMM;
  } else {
    EnterCriticalSection (&(((TComInfo *)dvid)->tpcs_m));

    if (err != NULL) *err = ERR_NON;
    result = _ReadPos (dvid, id, err);

    LeaveCriticalSection (&(((TComInfo *)dvid)->tpcs_m));
  }
  return result;
}

//------------------------------------------------------------------------------
//  bool KRS_WriteSyncData (TDeviceID dvid, uint8_t *dat, uint32_t size, TErrorCode *err)
//------------------------------------------------------------------------------
//  DESCRIPTION:
//    任意アドレスへ任意バイト数のデータ書き込み
//  PARAMS:
//    TDeviceID       dvid        デバイスID
//    uint8_t         *dat        adr + len + (ID+位置H+位置L)*n  (Dynamixelと等価)
//    uint32_t        size        書き込むデータのサイズを指定
//    TErrorCode      *err        エラーコード
//  RETURNS:
//    true            成功
//    false           エラー
//
bool KRSAPIDLL KRS_WriteSyncData (TDeviceID dvid, uint8_t *dat, uint32_t size, TErrorCode *err) {
  bool    result = false;
  uint8_t adr, len, idx;

  if (dvid == 0) {
    if (err != NULL) *err = ERR_INVALID_DEVID;
    return false;
  }
  if ((((TComInfo *)dvid)->hComm == INVALID_HANDLE_VALUE)) {
    if (err != NULL) *err = ERR_COMM;
    return false;
  }

  EnterCriticalSection (&(((TComInfo *)dvid)->tpcs_m));

  if (err != NULL) *err = ERR_NON;

  if (size > DXMAXLENGTH || size < 5) { // 最大バッファ以上又は1件分のデータ未満
    if (err != NULL) *err = ERR_ILLEGAL_SIZE;
  } else {
    adr = dat[0];
    len = dat[1];
    idx = 2;
    if ((((size - 2) / (len + 1)) -1) * 3 + 3 >= size) {  // 3Bytes単位
      if (err != NULL) *err = ERR_ILLEGAL_SIZE;
    } else {
      while (idx < size) {
        if (!KRS_WriteBlockData (dvid, dat[idx], adr, &dat[idx+1], len, err)) break;
        idx += (len + 1);
      }
      result = true;
    }
  }

  LeaveCriticalSection (&(((TComInfo *)dvid)->tpcs_m));

  return result;
}
